import controlP5.*;

import peasy.test.*;
import peasy.org.apache.commons.math.*;
import peasy.*;
import peasy.org.apache.commons.math.geometry.*;

import saito.objloader.*;

PFont createdFont;

PeasyCam cam;
OBJModel objmodel;

float importScale = 1000.0f;

HEModel hemodel;

World world;

ControlP5 cp5;
Textarea importLogArea;
Textarea logArea;
CheckBox drawCheck;

String log ="";
String importLog = "";

color solvedColor = color(180, 80, 130);
color optimalSolvedColor = color(80, 180, 180);
color funnelSolvedColor = color(200, 200, 100);

ArrayList<PVector> nodeIndexOnScreen;
void setup() {
  size(1400, 800, P3D);
  nodeIndexOnScreen = new ArrayList<PVector>();
  cam = new PeasyCam(this, 750);
  cam.setMinimumDistance(10);
  cam.setMaximumDistance(1500);
  cam.rotateX(PI/180.0 * 30.0 );
  cam.rotateY(PI/180.0 * -30.0 );

  createdFont = createFont("Arial", 32, true);
  textFont(createdFont);

  /* ---begin convertion .obj model to Half Edge Data Structure.--- */
  objmodel = new OBJModel(this, "Data/test.obj");
  objmodel.disableMaterial();
  objmodel.scale(importScale);
  objmodel.translate(new PVector(0, -1, 0));
  
  hemodel = new HEModel();
  //vertex
  importLog += "obj model vertex count:" + objmodel.getVertexCount () +"\n";
  for (int i=0; i < objmodel.getVertexCount (); i++) {
    PVector p = objmodel.getVertex(i);
    //HEVertex v = new HEVertex(p.x, p.y, p.z);
    HEVertex v = new HEVertex(p.x, p.y, p.z, i);
    hemodel.vertices.add(v);
  }
  //edge and loop
  for (int i=0; i < objmodel.getSegmentCount (); i++) {
    for (int j=0; j < objmodel.getFaceCount (); j++) {
      HEFace f = new HEFace();
      Face face = objmodel.getFaceInSegment(i, j);
      //Create HalfEdge
      for (int k = 0; k < face.getVertexCount (); k++) {
        HalfEdge e = new HalfEdge((HEVertex)hemodel.vertices.get(face.getVertexIndex(k)), f);
        hemodel.addEdge(e);
        f.addEdge(e);
      }
      //Set NextEdge to each HalfEdge
      for (int k=0; k < f.getEdgeCount (); k++) {
        int nextIndex = k+1;
        if (nextIndex > f.getEdgeCount()-1)nextIndex =0;
        f.getEdge(k).setNextEdge(f.getEdge(nextIndex));
        int prevIndex = k-1;
        if (prevIndex < 0)prevIndex =f.getEdgeCount()-1;
        f.getEdge(k).setPrevEdge(f.getEdge(prevIndex));
      }

      hemodel.addFace(f);
      f.index = hemodel.getFaceCount() - 1;
    }
  }
  importLog += "Face count:" + hemodel.faces.size() +"\n";

  //Set Opposite Edge
  for (int i=0; i < hemodel.getEdgeCount (); i++) {
    HalfEdge e = hemodel.getEdge(i);
    for (int j=0; j < hemodel.getEdgeCount (); j++) {
      HalfEdge t = hemodel.getEdge(j);
      if (e != t && e.oppositeEdge == null) {
        if (e.nextEdge.vertex == t.vertex && e.vertex == t.nextEdge.vertex) {
          e.setOppositeEdge(t);
        }
      }
    }
  }
  /* ---End of Convertion.--- */

  /* ---Generate waypoint data---*/
  world = new World();
  for (HEFace face : hemodel.faces) {
    Node n = new Node(face.getCenter());
    n.index = face.index;
    ArrayList<HEFace> linkedFaces = face.getLinkedFaces();
    int[] linkIndices = new int[linkedFaces.size()];
    for (int i = 0; i < linkedFaces.size (); i++) {
      linkIndices[i] = linkedFaces.get(i).index;
    }
    n.linkedNodeIndices = linkIndices;
    world.addNode(n);
  }
  importLog += "node count" + world.nodes.size() + "\n";

  //set start and goal
  int lastNodeIndex = world.nodeSize()-1;
  world.setStartNode(0);
  world.setGoalNode(lastNodeIndex);

  buildGUI();

  importLog += "waypoint data has generated.\n";
  importLogArea.setText(importLog);
}

void draw() {
  nodeIndexOnScreen.clear();
  log = "";
  log += "Start/Goal:" + world.startNode.index +"/" + world.goalNode.index + "\n";
  if (cp5.window(this).isMouseOver()) {
    cam.setActive(false);
  } else {
    cam.setActive(true);
  }

  world.update();

  //Store face on route
  ArrayList<HEFace> faceOnRoute;
  faceOnRoute = new ArrayList<HEFace>();

  for (Node node : world.path) {
    HEFace f = hemodel.faces.get(node.index);
    faceOnRoute.add(f);
  }
  //Store edge on route
  ArrayList<HalfEdge> edgeOnRoute = new ArrayList<HalfEdge>();
  for (Node node : world.path) {
    if (node.prev != null) {
      HEFace f = hemodel.faces.get(node.index);
      for (HalfEdge e : f.edges) {
        if (e.oppositeEdge != null) {
          if (e.oppositeEdge.face.index == node.prev.index) {
            edgeOnRoute.add(e);
          }
        }
      }
    }
  }

  background(80);
  if (drawCheck.getState("Axis"))drawAxis();
  if (drawCheck.getState("Mesh")) drawAllEdges();
  if (drawCheck.getState("NodeGraph"))drawLinks();
  drawStartAndGaol();
  drawRoute();

  drawFaceOnRoute(faceOnRoute);
  drawEdgeOnRoute(edgeOnRoute);

  ArrayList<PVector> optimalPath = getEdgeOptimizePath(world.goalNode.getVector(), world.startNode.getVector(), edgeOnRoute);
  strokeWeight(3);
  stroke(optimalSolvedColor);
  drawPath(optimalPath);

  
  //funnel optimize
  /*
  strokeWeight(1);
  ArrayList<PVector> funnelOptimalPath = getFunnelOptimizePath(world.goalNode.getVector(), world.startNode.getVector(), edgeOnRoute,cp5.get(Slider.class,"FunnelMargin").getValue());

  stroke(funnelSolvedColor);
  strokeWeight(2);
  //draw funnel optimized path
  for (int i=0; i < funnelOptimalPath.size ()-1; i++) {
    PVector v1 = funnelOptimalPath.get(i);
    PVector v2 = funnelOptimalPath.get(i+1);
    line(v1.x, v1.y, v1.z, v2.x, v2.y, v2.z);
    drawCross(v1, 10);
  }
  */
  logArea.setText(log);
  drawGUI();
}


