void drawAllEdges() {
  for (HalfEdge e : hemodel.edges) {
    HalfEdge next = e.nextEdge;
    if (e.isHalf()) {
      stroke(0);
      strokeWeight(1);
    } else {
      stroke(128);
      strokeWeight(1);
    }
    line(e.vertex.x, e.vertex.y, e.vertex.z, next.vertex.x, next.vertex.y, next.vertex.z);
  }
  strokeWeight(1);
}

void drawAxis() {
  float axisMag = 100;
  stroke(255, 0, 0);
  line(0, 0, 0, axisMag, 0, 0);
  stroke(0, 255, 0);
  line(0, 0, 0, 0, axisMag, 0);
  stroke(0, 0, 255);
  line(0, 0, 0, 0, 0, axisMag);
}


void drawCross(PVector center, float size) {
  line(center.x - size/2.0, center.y, center.z - size/2.0, center.x + size/2.0, center.y, center.z + size/2.0);
  line(center.x - size/2.0, center.y, center.z + size/2.0, center.x + size/2.0, center.y, center.z - size/2.0);
}


void drawLinks() {
  for (HEFace l : hemodel.faces) {
    PVector center = l.getCenter();
    stroke(0);
    strokeWeight(2);
    drawCross(center, 5);

    ArrayList<HEFace> linkedfaces = l.getLinkedFaces();
    strokeWeight(1);
    stroke(100, 180, 100);
    for (HEFace link : linkedfaces) {
      line(l.getCenter().x, l.getCenter().y, l.getCenter().z, link.getCenter().x, link.getCenter().y, link.getCenter().z );
    }
    strokeWeight(1);
  }

  for (HEFace f : hemodel.faces) {
    nodeIndexOnScreen.add(new PVector(screenX(f.getCenter().x, f.getCenter().y, f.getCenter().z),screenY(f.getCenter().x, f.getCenter().y, f.getCenter().z)));
  }
}

void drawStartAndGaol() {
  strokeWeight(5);
  stroke(255, 0, 0);
  drawCross(new PVector(world.startNode.position.x, world.startNode.position.y - 1, world.startNode.position.z), 10);
  stroke(0, 0, 255);
  drawCross(new PVector(world.goalNode.position.x, world.goalNode.position.y - 1, world.goalNode.position.z), 10);
  strokeWeight(1);
}

void drawRoute() {
  stroke(solvedColor);  
  for (Node node : world.path) {
    if (node.prev != null) {
      strokeWeight(3);
      stroke(solvedColor);
      line(node.position.x, node.position.y, node.position.z, node.prev.position.x, node.prev.position.y, node.prev.position.z);
    }
  }
}

void drawFaceOnRoute(ArrayList<HEFace> faceOnRoute) {
  strokeWeight(1);
  stroke(50, 50, 90);
  for (HEFace f : faceOnRoute) {
    //draw
    for (HalfEdge e : f.edges) {
      HalfEdge next = e.nextEdge;
      PVector se = e.vertex.getVector();
      se.sub(e.face.getCenter());
      se.mult(0.8);
      se.add(e.face.getCenter());
      PVector sn = next.vertex.getVector();
      sn.sub(next.face.getCenter());
      sn.mult(0.8);
      sn.add(next.face.getCenter());
      line(se.x, se.y, se.z, sn.x, sn.y, sn.z);
    }
  }
}

void drawEdgeOnRoute(ArrayList<HalfEdge> edgeOnRoute) {
  stroke(optimalSolvedColor);
  for (HalfEdge e : edgeOnRoute) {
    strokeWeight(1);
    line(e.vertex.x, e.vertex.y, e.vertex.z, e.nextEdge.vertex.x, e.nextEdge.vertex.y, e.nextEdge.vertex.z);
    strokeWeight(2);
    drawCross(e.vertex.getVector(), 6);
  }
}

void drawPath(ArrayList<PVector> path) {
  for (int i=0; i < path.size ()-1; i++) {
    line(path.get(i).x, path.get(i).y, path.get(i).z, path.get(i+1).x, path.get(i+1).y, path.get(i+1).z);
  }
}

