class World {
  private ArrayList<Node> nodes;
  private ArrayList<Node> openNodes;
  private ArrayList<Node> path;
  Node startNode, goalNode;

  World() {
    nodes = new ArrayList<Node>();
    openNodes = new ArrayList<Node>();
    path = new ArrayList<Node>();
    //setStartNode(0);
    //setGoalNode(nodeSize()-1);
  }
  World(int numOfNode) {
    nodes = new ArrayList<Node>();
    openNodes = new ArrayList<Node>();
    for (int i = 0; i < numOfNode; i++) {
      nodes.add(new Node(new PVector(random(0, width), random(0, height))));
    }
    setStartNode(0);
    setGoalNode(nodeSize()-1);
  }

  World(int numX, int numY) {
    nodes = new ArrayList<Node>();
    openNodes = new ArrayList<Node>();
    for (int j =0; j < numY; j++) {
      for (int i = 0; i < numX; i++) {
        nodes.add(new Node(new PVector(i*(width-50)/(numX-1) + 25, j*(height-50)/(numY-1) + 25)));
      }
    }
    setStartNode(0);
    setGoalNode(nodeSize()-1);
  }

  void update() {
    resetNodes();
    
    solveStart();

    //detect route
    for (int i=0; i < nodeSize (); i++) {
      Node n = getNode(i);
      if (n.nodeType == NodeType.GOAL && n.done) {
        n.isRoute = true;
        path.add(n);
        Node p = n.prev;
        while (p != null) {
          p.isRoute = true;
          path.add(p);
          p = p.prev;
        }
      }
    }
  }

  public Node getNode(int index) {
    return nodes.get(index);
  }

  public void removeNode(int index) {
    if (nodes.size() > 2) {
      nodes.remove(index);
    }
  }

  public void addNode(Node node) {
    nodes.add(node);
  }

  public int nodeSize() {
    return nodes.size();
  }

  private void resetNodes() {
    openNodes.clear();
    path.clear();
    for (int i = 0; i < nodeSize (); i++) {
      getNode(i).reset();
    }
  }

  private float heuristicAt(PVector position) {
    return position.dist(goalNode.position);
  }

  private void solveStart() {
    startNode.score = 0;
    startNode.fs = startNode.score + heuristicAt(startNode.position);
    openNodes.add(startNode);
    solveNext();
  }

  void solveNext() {
    float minScore = Float.MAX_VALUE;
    int minIndex = -1;
    Node u = null;
    for (int i = 0; i < openNodes.size (); i++) {
      Node targetNode = openNodes.get(i);
      if (targetNode.done) continue;
      if (targetNode.fs < minScore) {
        minScore = targetNode.fs;
        minIndex = i;
        u = targetNode;
      }
    }

    if (u == null) {
      return;
    }

    openNodes.remove(minIndex);
    u.done = true;

    if (u.nodeType == NodeType.GOAL) {
      return;
    }

    int[] linkNodes = u.linkedNodeIndices;
    for (int linkIndex :linkNodes) {
      Node v = getNode(linkIndex);
      if (v.done || v.nodeType == NodeType.OBSTACLE) continue;
      //following number 1 is moving cost between node to node.
      float cost = dist(u.position.x, u.position.y, v.position.x, v.position.y);
      if (u.score + cost + heuristicAt(v.position) < v.fs) {
        v.score = u.score + cost;
        v.fs = v.score + heuristicAt(v.position);
        v.prev = u;

        if (openNodes.indexOf(v) == -1)openNodes.add(v);
      }
    }
    solveNext();
  }

  public int setStartNode(int index) {
    if(index < 0){
      index = 0;
    }
    if(index >= nodes.size()){
      index = nodes.size() -1;
    }
    if (getNode(index).nodeType == NodeType.START || getNode(index).nodeType == NodeType.GOAL) return -1;
    if (startNode != null) {
      startNode.nodeType = NodeType.NONE;
    }
    getNode(index).nodeType = NodeType.START;
    startNode = getNode(index);
    return index;
  }

  public int setGoalNode(int index) {
    if(index < 0){
      index = 0;
    }
    if(index >= nodes.size()){
      index = nodes.size() -1;
    }
    if (getNode(index).nodeType == NodeType.START || getNode(index).nodeType == NodeType.GOAL) return -1;
    if (goalNode != null) {
      goalNode.nodeType = NodeType.NONE;
    }
    getNode(index).nodeType = NodeType.GOAL;
    goalNode = getNode(index);
    return index;
  }

  public void setObstacleNode(int index) {
    Node n = getNode(index);
    if (n.nodeType != NodeType.START && n.nodeType != NodeType.GOAL) {
      if (n.nodeType == NodeType.NONE) {
        n.nodeType = NodeType.OBSTACLE;
      } else if (n.nodeType == NodeType.OBSTACLE) {
        n.nodeType = NodeType.NONE;
      }
    }
  }
}

